# Autogenerated by Astropy's setup.py on 2019-06-10 21:59:18 UTC
from __future__ import unicode_literals
import datetime

version = "3.2"
githash = "7944b5b7ae4f13079519fff0d4f9e5f7e7ee4b01"


major = 3
minor = 2
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 6, 10, 21, 59, 18)
debug = False

astropy_helpers_version = "3.2"
